{-# LANGUAGE LambdaCase, TypeSynonymInstances, FlexibleInstances, MultiParamTypeClasses, OverloadedStrings #-}
{-# LANGUAGE DuplicateRecordFields #-}

module Main where

import GHC.Generics (Generic)
import Data.Aeson   (FromJSON)
import Data.Binary  (Binary)
import Control.Monad ((>=>))
import System.FilePath
import Data.Text (Text)
import Data.List (intersperse)
import Lucid
import Data.Yaml
import Control.Applicative ((<|>))

import Achille
import Achille.Writable (Writable)
import qualified Achille.Writable as Writable
import Achille.Internal.IO (AchilleIO)

import Achille.Task.Pandoc

import Data.Functor
import Config (config, ropts, wopts, SiteConfig(title))
import Text.Pandoc.Options


-- Bibliography info

data Author = Author
  { name  :: Text
  , url   :: Maybe Text
  , orcid :: Maybe Text
  } deriving (Eq, Show, Generic, Binary)

instance FromJSON Author where
  parseJSON o =
    withObject "Author info"
      (\v -> Author <$> v .: "name"
                    <*> v .:? "url"
                    <*> v .:? "orcid") o
    <|> withText "Author name" (\n -> pure $ Author n Nothing Nothing) o

data Venue = Venue
  { name :: Text
  , url :: Text
  } deriving (Eq, Show, Generic, Binary, FromJSON)

data Publication = Publication
  { title   :: Text
  , slug    :: Text
  , authors :: [Author]
  , file    :: Maybe Text
  , doi     :: Maybe Text
  , venue   :: Venue
  } deriving (Eq, Show, Generic, Binary, FromJSON)

parseYaml :: (AchilleIO m, FromJSON a) => FilePath -> Task m a
parseYaml p = do
  res <- decodeEither' <$> readBS p
  case res of
    Left err -> fail (prettyPrintParseException err)
    Right ok -> return ok


-- compile markdown with custom options and extensions
compileMD = compilePandocWith def { readerExtensions = pandocExtensions } def

-- writing Html to disk (efficiently)
instance AchilleIO m => Writable m (Html ()) where
    write to = Writable.write to . renderBS


main = achille do
  match_ "assets/*" $ copyFile
  match_ "static/*" $ copyFileAs (makeRelative "static/")
  match_ "papers/*" $ copyFile

  summary <- matchFile "summary.md" compileMD
  pubs :: [Publication] <- matchFile "publications.yaml" parseYaml

  watch summary $ watch pubs $ match_ "index.md" \src -> do
     compileMD src
      <&> renderIndex summary (renderPublications pubs)
      >>= write (src -<.> "html")

-- TODO: fix year
renderPublications :: [Publication] -> Html ()
renderPublications pubs = section_ [ id_ "publications" ] do
  h2_ "Publications"
  ul_ [ data_ "year" "2022" ] $ foldMap renderPub pubs
  where renderPub :: Publication -> Html ()
        renderPub Publication {..} = li_ [id_ slug] do
          p_ [class_ "title"] $ a_ [href_ ("#" <> slug)] $ toHtmlRaw title
          p_ [class_ "authors"] $ renderAuthors (reverse (fmap renderAuthor authors))
          let Venue vName vUrl = venue
          p_ [class_ "venue"] $ a_ [href_ vUrl] $ toHtmlRaw vName
          p_ [class_ "buttons"] do
            maybe mempty (\url -> a_ [href_ ("/" <> url)] "PDF") file
            maybe mempty (\doi -> a_ [href_ ("https://doi.org/" <> doi)] "DOI") doi

        renderAuthor :: Author -> Html ()
        renderAuthor (Author name url orcid) =
          case url of
            Nothing  -> toHtmlRaw name
            Just url -> a_ [href_ url] (toHtmlRaw name)

        renderAuthors :: [Html ()] -> Html ()
        renderAuthors [] = mempty
        renderAuthors [one] = one
        renderAuthors (last:others) =
          mconcat (intersperse ", " (reverse others)) <> " and " <> last

renderIndex :: Text -> Html () -> Text -> Html ()
renderIndex summary pubs body =
  doctypehtml_ do
    head_ do
      meta_ [ name_ "viewport"
            , content_ "width=device-width, initial-scale=1.0, user-scalable=yes"
            ]
      meta_ [ name_ "theme-color",  content_ "#000000" ]
      meta_ [ name_ "robots", content_ "index, follow" ]
      meta_ [ charset_ "utf-8" ]
      link_ [ rel_ "stylesheet",  href_ "/assets/theme.css" ]
      link_ [ rel_  "shortcut icon"
            , type_ "image/svg"
            , href_ "/assets/favicon.svg"
            ]
      -- meta_ [ property_ "og:title", content_ title ]
      -- meta_ [ property_ "og:type", content_ "website" ]
      -- meta_ [ property_ "og:image", content_ image ]
      -- meta_ [ property_ "og:description", content_ description ]
      title_ "Lucas Escot"

    body_ do
      aside_ [id_ "summary"] $ toHtmlRaw summary
      main_ do
        section_ $ toHtmlRaw body
        pubs
      footer_ do
        "2021 · "
        a_ [ href_ "https://creativecommons.org/licenses/by-nc/2.0/" ]
           "CC BY-NC 2.0"
